from pandas import DataFrame, merge, concat, \
    to_timedelta
from offline_results.common.config import DURATION_CUTOFF, \
    BUCKET_NAME, USER_MAPPING_KEY, PREF_FEATURES
from offline_results.common.constants import CUSTOMER_ID, VIDEO_ID1, \
    DURATION, CSV_EXTENSION, USER_LABEL, RELATIONSHIP, ACTOR, \
    RATING, ATTRIBUTE1, MESSAGE
from offline_results.preferences.process_generate_pref import Preferences
from graphdb.graph import GraphDb
from offline_results.common.read_write_from_s3 import ConnectS3
from offline_results.preferences.relationship_extractor import \
    RelationshipExtractor
from datetime import datetime
import pandas as pd


class PreferenceUpdater:

    def __init__(self, ubd_new, connection_object):
        self.duration_cutoff = DURATION_CUTOFF
        self.ubd_new = ubd_new
        self.graph = GraphDb.from_connection(connection_object)

    def find_cutoff_users(
            self,
    ) -> DataFrame:
        """
        Find customer_ids that have duration above
        specified threshold after grouping.
        :return: Dataframe object pandas
        """
        data = self.ubd_new
        data[CUSTOMER_ID] = data[CUSTOMER_ID].astype(str)
        data = data[data[CUSTOMER_ID] != "0"]
        data = data[data.created_on > datetime.now() - to_timedelta("8 days")]
        data = data.groupby([CUSTOMER_ID, VIDEO_ID1])[DURATION].sum().reset_index()
        data = data[data[DURATION] > self.duration_cutoff].reset_index(drop=True)
        data = data.drop(columns=[DURATION])
        data = merge(self.ubd_new, data, on=[CUSTOMER_ID, VIDEO_ID1], how='inner')

        return data

    def get_preferences(
            self,
            data: DataFrame,
    ) -> DataFrame:
        """
        Find user preferences for the considered users
        :param data: dataframe object pandas
        :return: preferences dataframe object pandas
        """
        mi = Preferences(data)
        pref_df = mi.controller()

        return pref_df

    def update_user_preference(
            self,
            feature: str,
            s3_data: DataFrame,
            new_data: DataFrame
    ) -> DataFrame:
        """
        Find user preferences for the considered users
        :param feature: feature name to update
        :param s3_data: dataframe object pandas
        :param new_data: dataframe object pandas
        :return: preferences dataframe object pandas
        """
        s3_data[CUSTOMER_ID] = s3_data[CUSTOMER_ID].astype(str)
        existing_user = new_data[new_data[CUSTOMER_ID].isin(s3_data[CUSTOMER_ID])]
        new_user = new_data[~new_data[CUSTOMER_ID].isin(s3_data[CUSTOMER_ID])]
        cols_to_list = [col for col in new_data if col.startswith(feature)]
        cols_to_add = list(set(cols_to_list) - set(s3_data.columns.to_list()))
        cols_to_list.append(CUSTOMER_ID)
        cols_to_add.append(CUSTOMER_ID)
        existing_user = existing_user[cols_to_list]
        new_user = new_user[cols_to_list]

        if len(cols_to_add) > 1:
            s3_data = s3_data.merge(existing_user[cols_to_add], on=CUSTOMER_ID, how='left')
            s3_data = s3_data.set_index(CUSTOMER_ID)
            existing_user = existing_user.set_index(CUSTOMER_ID)
            s3_data.update(existing_user.drop(columns=cols_to_add[:-1]))
            s3_data = s3_data.reset_index().rename(columns={'index': CUSTOMER_ID})

        if not new_user.empty:
            s3_data = concat([s3_data, new_user], ignore_index=True)

        s3_data = s3_data.fillna(0)

        return s3_data.reset_index(drop=True)

    def fetch_dump_preferences(
            self,
            data: DataFrame,
            bucket_name=None,
            object_name=None,
            resource=None,
    ):
        """
        Fetches the preference feature files from s3 and
        passes to another function to update preferences.
        :param data: dataframe of new feature preference
        :param bucket_name: s3 bucket name
        :param object_name: s3 object name
        :param resource: boto connection resource
        """
        for feature, feature_relation in PREF_FEATURES.items():
            feature_pref = ConnectS3().read_csv_from_s3(bucket_name=BUCKET_NAME,
                                                        object_name=USER_MAPPING_KEY + feature + CSV_EXTENSION,
                                                        resource=resource)

            updated_preferences = self.update_user_preference(feature=feature,
                                                              s3_data=feature_pref,
                                                              new_data=data)

            print("Saving updated raw preferences in S3...")
            ConnectS3().write_csv_to_s3(bucket_name=BUCKET_NAME,
                                        object_name=USER_MAPPING_KEY + feature + CSV_EXTENSION,
                                        df_to_upload=updated_preferences,
                                        resource=resource)

            to_dump_rel = self.filter_users(all_users=updated_preferences,
                                            users_to_keep=data[CUSTOMER_ID].tolist())

            to_dump_rel = self.drop_nan_features(data=to_dump_rel)

            to_dump_rel = self.get_feature_relations(data=to_dump_rel)

            self.drop_existing_pref_rel(data=data[[CUSTOMER_ID]],
                                        relation_label=feature_relation[0]
                                        )
            print("Starting dumping relationships...")
            self.dump_relations(dump_data=to_dump_rel,
                                feature=feature,
                                relation_label=feature_relation[0],
                                dest_prop=feature_relation[1]
                                )

    def get_feature_relations(
            self,
            data: DataFrame
    ) -> DataFrame:
        """
        Use the relationship extractor to extract
        relationship dataframe object based on
        user's preferences
        :param data: dataframe object pandas
        :return: dataframe object pandas with each
        record representing a relationship between
        a user and a node type.
        """
        re = RelationshipExtractor(
            data=data
        )
        return re.controller()

    def filter_users(
            self,
            all_users: DataFrame,
            users_to_keep: list
    ) -> DataFrame:
        """
        Filter users to update preferences
        :param all_users: Dataframe object pandas
        :param users_to_keep: list of customer_ids
        :return : Dataframe object pandas
        """
        all_users[CUSTOMER_ID] = all_users[CUSTOMER_ID].astype(str)
        return all_users[all_users[CUSTOMER_ID].isin(users_to_keep)] \
            .reset_index(drop=True)

    def drop_nan_features(
            self,
            data
    ) -> DataFrame:
        """
        Drop nan columns from dataframe
        :param data: Dataframe object pandas
        :return : Dataframe object pandas
        """
        to_drop = []
        for feature in data.columns:
            if feature[len(feature) - 4:] == '_nan':
                to_drop.append(feature)
        return data.drop(columns=to_drop)

    def drop_existing_pref_rel(
            self,
            data,
            relation_label,
    ):
        """
        Function to drop edge from user node
        :param data: dataframe object pandas
        :param relation_label: name of edge
        """
        for index in range(len(data)):
            print("Dropping existing relationship {} with user {}".
                  format(relation_label, data.loc[index, CUSTOMER_ID]))
            self.graph.custom_query(
                query=f'''g.V().hasLabel('{USER_LABEL}').
                has('{CUSTOMER_ID}', '{data.loc[index, CUSTOMER_ID]}').
                outE('{relation_label}').drop()'''
            )

    def dump_relations(
            self,
            dump_data: DataFrame,
            feature: str,
            relation_label: str,
            dest_prop: str,
    ):
        """
        Function to dump user preferences in graphdb
        using custom query.
        :param dump_data: preference dataframe to be dumped
        :param feature: label of destination node
        :param relation_label: name of edge
        :param dest_prop: property of destination node
        """
        undumped_rel = DataFrame()
        dest_label = feature

        if relation_label in ["HAS_ACTOR_PREFERENCE", "HAS_DIRECTOR_PREFERENCE"]:
            dest_label = ACTOR

        for index in range(len(dump_data)):
            print("Dumping relation {} of {}".format(index + 1, len(dump_data)))

            try:
                source_node = self.graph.custom_query(
                    query=f'''g.V().hasLabel('{USER_LABEL}').
                    has('{CUSTOMER_ID}', '{dump_data.loc[index, CUSTOMER_ID]}')'''
                )[0][0].id

                if feature in [RATING, ATTRIBUTE1]:
                    destination_node = self.graph.custom_query(
                        query=f'''g.V().hasLabel('{feature}').
                        has('{dest_prop}', '{dump_data.loc[index, dump_data.columns[-1]]}')'''
                    )[0][0].id
                else:
                    destination_node = self.graph.custom_query(
                        query=f'''g.V().hasLabel('{dest_label}').
                        has('{dest_prop}', {dump_data.loc[index, dump_data.columns[-1]]})'''
                    )[0][0].id

                print("Creating relationship between customer_id:{} and {}:{}".
                      format(dump_data.loc[index, CUSTOMER_ID], feature,
                             dump_data.loc[index, dump_data.columns[-1]])
                      )

                self.graph.custom_query(
                    query=f'''g.V('{source_node}').as('v').V('{destination_node}').
                    coalesce(__.inE('{relation_label}').where(outV().as('v')),
                    addE('{relation_label}').from('v'))'''
                )

            except Exception as e:
                print("Unable to create {} relation between customer_id:{} and {}:{}".
                      format(relation_label, dump_data.loc[index, CUSTOMER_ID],
                             feature, dump_data.loc[index, dump_data.columns[-1]])
                      )
                error_dumping = DataFrame.from_dict({CUSTOMER_ID: [dump_data.loc[index, CUSTOMER_ID]],
                                                     RELATIONSHIP: relation_label,
                                                     feature: [dump_data.loc[index, dump_data.columns[-1]]],
                                                     MESSAGE: e})
                undumped_rel = pd.concat([undumped_rel, error_dumping], axis=0, ignore_index=True)

        undumped_rel.to_csv("undumped_" + feature + "_relations.csv", index=False)

    def controller(
            self,
            bucket_name=None,
            object_name=None,
            resource=None,
    ):
        """
        Driver function for PreferenceUpdater class.
        :param bucket_name: s3 bucket name
        :param object_name: s3 object name
        :param resource: boto connection resource
        """
        data = self.find_cutoff_users()
        data = self.get_preferences(data)
        self.fetch_dump_preferences(data=data,
                                    bucket_name=bucket_name,
                                    object_name=object_name,
                                    resource=resource,
                                    )

        return data
