from datetime import datetime
from typing import ClassVar

import pandas as pd
import numpy as np
from graphdb import GraphDb, GraphDbConnection
from offline_results.recommendation.homepage_recommendation.because_you_watched.prepare_byw_data import BYWData
from functools import reduce
from offline_results.common.constants import CUSTOMER_ID, RECOMMENDATION_SCORE, RECOMMENDED_CONTENT, RC_CREATED_ON, \
     PAY_TV_CONTENT, NO_PAY_TV_CONTENT


class BYWModel:
    def __init__(
            self,
            connection_class: GraphDbConnection
    ):
        self.graph = GraphDb.from_connection(connection_class)

    @classmethod
    def from_connection_uri(
            cls,
            connection_uri: str
    ) -> ClassVar:
        """Create new object based on connection uri
        :param connection_uri: string connection uri
        :return: object class
        """
        return cls(GraphDbConnection.from_uri(connection_uri, connection_uri))

    @classmethod
    def from_connection_class(
            cls,
            connection_class: GraphDbConnection
    ) -> ClassVar:
        """Define new class based on object connection
        :param connection_class: object connection class
        :return: object class
        """
        return cls(connection_class)

    @staticmethod
    def byw_algorithm(pay_tv_content: bool = False, graph=GraphDbConnection):
        if pay_tv_content:
            content_label = PAY_TV_CONTENT
        else:
            content_label = NO_PAY_TV_CONTENT
        content_similarity_data = BYWData.get_all_content_similarity(content_label, graph)
        user_history_data = BYWData.get_viewed_content(graph)
        all_recommendation_dict = {}
        print("Generating recommended content to the all customer")
        print("====================================================")
        for idx, data_property in user_history_data.iterrows():
            list_content_id = data_property.content_id
            customer_id = data_property.customer_id
            if len(list_content_id) == 1:
                # USE-CASE CONDITION : Recommending the content if customer only watched 1 content

                # Iterate through all content_id which watched by user
                for id_content in list_content_id:
                    # Iterate through all content similarity data
                    for key_content_id, similarity_data in content_similarity_data.items():
                        # If the content which watched by user is available on the content_similarity_data
                        if id_content == key_content_id:
                            # Take the first 10 most similar content from similarity_data
                            list_recommended_content_id = []
                            for recommended_content_id, sim_score in similarity_data.items():
                                list_recommended_content_id.append(recommended_content_id)
                                if len(list_recommended_content_id) == 10:
                                    break
                            # Create dictionary of user-recommended content
                            all_recommendation_dict[customer_id] = list_recommended_content_id
                            break
            elif len(list_content_id) > 1:
                # USE-CASE CONDITION : Recommending the content if customer watched multiple content

                list_df_recommendation = []
                # Iterate through all content_id which watched by user
                id_content_position = idx + 1
                for id_content in list_content_id:
                    # Iterate through all content similarity data
                    for key_content_id, similarity_data in content_similarity_data.items():
                        # If the content which watched by user is available on the content_similarity_data
                        if id_content == key_content_id:
                            list_recommended_content_id = []
                            # Take the first 10 most similar content from similarity_data
                            for recommended_content_id, sim_score in similarity_data.items():
                                list_recommended_content_id.append(recommended_content_id)
                                # When length of the list already has 10 content_id data
                                if len(list_recommended_content_id) == 10:
                                    # Give the point recommendation_score to all recommended content_id
                                    # The most similar content_id will give 10 points divided by id_content_position
                                    # The Most Similar Content Point = (10/id_content_position)

                                    # Until the least similar content_id will give 1 point divided by id_content_position
                                    # The Least Similar Content Point = (10/id_content_position)

                                    dict_recommendation = dict(zip(list_recommended_content_id,
                                                                   np.arange(10/id_content_position,
                                                                             0/id_content_position,
                                                                             -1/id_content_position)))

                                    # convert to dataframe and sort by recommendation score in descending
                                    df_temp = pd.DataFrame({RECOMMENDATION_SCORE: dict_recommendation})\
                                        .sort_values(by=RECOMMENDATION_SCORE, ascending=False)
                                    # append df_temp recommendation from current content_id to the list_df_recommendation
                                    list_df_recommendation.append(df_temp)
                                    id_content_position += 1
                                    break
                # After having multiple df_recommendation, we sum all recommendation_score points based on content_id
                final_df = reduce(lambda df_source, df_target: df_source.add(df_target, fill_value=0),
                                  list_df_recommendation)
                # After do summation from all dataframe, we remove content_id which already watched by user from
                # recommendation content dataframe
                for content_id in list_content_id:
                    if content_id in final_df.index.values.tolist():
                        final_df = final_df.drop(content_id)

                # Sort the recommendation content based on recommendation_score in descending method
                # and take the first 10 content with the highest score
                final_df = final_df.sort_values(by=RECOMMENDATION_SCORE, ascending=False).head(10)
                # Take the content_id from the index and transform it to the list
                final_recommended_content_id = final_df.index.values.tolist()
                # Create dictionary of user-recommended content
                all_recommendation_dict[customer_id] = final_recommended_content_id

        # This step is to create output as dataframe
        # which contain customer_id, recommended_content_id, and rc_created_on
        all_recommendation_dict_df = pd.DataFrame([all_recommendation_dict]).T.reset_index(level=0)
        all_recommendation_dict_df.columns = [CUSTOMER_ID, RECOMMENDED_CONTENT]
        all_recommendation_dict_df[RC_CREATED_ON] = datetime.today().strftime("%Y-%m-%d %H:%M:%S")

        return all_recommendation_dict_df
