from offline_results.common.config import MINIMUM_SIMILAR_USERS_THRESHOLD
from offline_results.common.constants import CONTENT_ID, VIEW_COUNT, IS_PAYTV
from offline_results.recommendation.homepage_recommendation.core.with_cf_user_to_user.utils \
    import UserViewershipUtils
from offline_results.similarity.user_profile.config import DEFAULT_CLUSTER_LABEL
import logging
logging.basicConfig(level=logging.INFO)

class UserViewershipController(UserViewershipUtils):

    def __init__(
            self,
            connection_object,
            s3_resource,
            s3_bucket_name: str,
            s3_object_name: str
    ):

        UserViewershipUtils.__init__(
            self,
            connection_object=connection_object,
            s3_resource=s3_resource,
            s3_bucket_name=s3_bucket_name,
            s3_object_name=s3_object_name
        )

    def controller(
            self,
            customer_id: str
    ):
        """
        The driver function to generate user-user CF based recommendation results. 
        The results are returned in the form of a dataframe object containing all 
        the relevant content and homepage related information. The entire procedure 
        includes the following steps:
        
        1) Accept a customer_id value from the parameter and identify the cluster_id 
        and paytv status of its user-profile cluster.
        
        2) Validate cluster_id
        
            a) If the user belongs to the default cluster, return an empty result set. 
            This is so because all users belonging to the default cluster are likely 
            to have no UBD records, thus making it impossible to find out the most 
            viewed contents from their user behaviour data.
            
            b) Else, find all the similar users to the passed customer_id within the 
            cluster_id cluster.
            
                i) If the number of similar users obtained are less than the specified 
                threshold, assume the complete set of users belonging to the cluster 
                {cluster_id, is_paytv} as the set of similar users to customer_id user.
                
        3) Fetch the viewership for each user in the set of similar users, and append 
        each result into a common dataframe object. The process is done paytv-wise so 
        that the paytv status can also be preserved to be returned along with the result.
        
        4) Validate viewership
        
            a) If no viewership records, return an empty set.
            
            b) Else, identify the following pieces of information
            
                i) Content-wise total view count
                ii) Content-wise paytv status (from step-3)
                iii) Merge i) and ii) with information computed during UserViewershipUtils 
                constructor initialisation to prepare the final result set.
                
        @param customer_id: customer id string value
        @return: Dataframe object pandas
        """
        
        # Accept a customer_id value from the parameter and identify the cluster_id 
        # and paytv status of its user-profile cluster.
        cluster_id, is_paytv = \
            self.get_cluster_id_for_user(
                customer_id=customer_id
            )

        # Validate cluster_id
        if cluster_id == DEFAULT_CLUSTER_LABEL:
            # If the user belongs to the default cluster, return an empty result set. 
            # This is so because all users belonging to the default cluster are likely 
            # to have no UBD records, thus making it impossible to find out the most 
            # viewed contents from their user behaviour data.
            logging.info("The user " +
                         customer_id +
                         " belongs to the " +
                         "default cluster. " +
                         "Unable to fetch results!")
            return None

        # Else, find all the similar users to the passed customer_id within the 
        # cluster_id cluster.
        similar_users = \
            self.get_similar_users(
                customer_id=customer_id,
                cluster_id=cluster_id,
                is_paytv=is_paytv
            )

        #  If the number of similar users obtained are less than the specified 
        #  threshold, assume the complete set of users belonging to the cluster 
        #  {cluster_id, is_paytv} as the set of similar users to customer_id user.
        if len(similar_users) < \
                MINIMUM_SIMILAR_USERS_THRESHOLD:
            logging.info("Number of similar users less " +
                         "than the specified threshold." +
                         "Retrieving all users in the same " +
                         "user profile cluster as of this user")
            similar_users = self.get_all_users(
                cluster_id=cluster_id,
                is_paytv=is_paytv
            )

        # Fetch the viewership for each user in the set of similar users,
        # and append each result into a common dataframe object. The process
        # is done paytv-wise so that the paytv status can also be preserved
        # to be returned along with the result.
        viewership_result = self.get_similar_users_viewership(
            similar_users=similar_users
        )

        # Validate viewership
        if len(viewership_result) == 0:
            # If no viewership records, return a null set.
            logging.info("No viewership record for " +
                         "the users similar to " +
                         customer_id +
                         " .Unable to fetch results!")
            return None

        # computing content-wise total view count
        content_view_counts = \
            viewership_result.groupby(
                CONTENT_ID
            )[VIEW_COUNT].size().reset_index()

        # computing content-wise paytv status (from step-3)
        content_paytv_type = \
            viewership_result[
                [CONTENT_ID, IS_PAYTV]
            ].drop_duplicates().reset_index(drop=True)

        # Merge above two dataframe objects with information computed
        # during UserViewershipUtils constructor initialisation to
        # prepare the final result set.
        return self.format_final_result(
            content_paytv_type,
            content_view_counts
        )
