import argparse
import ast
import random

import numpy as np
# Train and Test filename
import pandas as pd

TRAIN_FILENAME = "train_result.csv"
TEST_FILENAME = "test_result.csv"


class HomepageRecommendation:
    """
        A softmax classifier
    """
    __slots__ = ("epochs", "learning_rate", "batch_size", "regularization_strength", "wt", "momentum", "velocity")

    def __init__(self, epochs, learning_rate, batch_size, regularization_strength, momentum):
        """
        Softmax constructor which initialized parameters
        :param epochs: Number of iterations over complete training data
        :param learning_rate: A step size or a learning rate
        :param batch_size: A mini-batch size(less than total number of training data)
        :param regularization_strength: A regularization strength
        :param momentum: A momentum value
        """
        self.epochs = epochs
        self.learning_rate = learning_rate
        self.batch_size = batch_size
        self.regularization_strength = regularization_strength
        self.momentum = momentum
        self.velocity = None
        self.wt = None

    def train(self, x_train, y_train, x_test, y_test):
        """
        Train a softmax classifier model on training data using stochastic gradient descent with mini-batches
        and momentum to minimize softmax (cross-entropy) loss of this single layer neural network. It calcualtes
        mean per-class accuracy for the training/testing data and the loss.
        :param x_train: Training input data
        :param y_train: Training labels
        :param x_test: Testing input data
        :param y_test: Testing labels
        :return: A tuple of training/Testing losses and Accuracy
        """
        D = x_train.shape  # dimensionality
        label = np.unique(y_train)
        num_of_classes = len(label)  # number of classes
        y_train_enc = self.one_hot_encoding(y_train, num_of_classes)
        y_test_enc = self.one_hot_encoding(y_test, num_of_classes)
        self.wt = np.random.normal(0, 0.01, (D[0], num_of_classes))
        self.velocity = np.zeros(num_of_classes)


        train_losses = []
        test_losses = []
        train_acc = []
        test_acc = []
        for e in range(self.epochs):  # loop over epochs
            train_loss = self.sgd_with_momentum(x_train, y_train_enc)
            test_loss, dw = self.compute_loss(x_test, y_test_enc)
            train_acc.append(self.mean_accuracy(x_train, y_train))
            test_acc.append(self.mean_accuracy(x_test, y_test))
            train_losses.append(train_loss)
            test_losses.append(test_loss)
            print("{:d}\t->\tTrainL : {:.7f}\t|\tTestL : {:.7f}\t|\tTrainAcc : {:.7f}\t|\tTestAcc: {:.7f}"
                  .format(e, train_loss, test_loss, train_acc[-1], test_acc[-1]))

        return train_losses, test_losses, train_acc, test_acc

    def sgd_with_momentum(self, x, y):
        """
        Stochastic gradient descent with mini-batches. It divides training data into mini-batches
        and compute loss and grad on that mini-batches and updates the weights. It repeats for all samples.
        :param x: An input samples
        :param y: An input labels
        :return: Total loss computed
        """
        losses = []
        # Randomly juggle up the data.
        random_indices = random.sample(range(x.shape[0]), x.shape[0])
        x = x[random_indices]
        y = y[random_indices]
        for i in range(0, x.shape[0], int(self.batch_size)):
            x_batch = x[i:i + int(self.batch_size)]
            y_batch = y[i:i + int(self.batch_size)]
            loss, dw = self.compute_loss(x_batch, y_batch)
            self.velocity = (self.momentum * self.velocity) + (
                    self.learning_rate * dw)
            self.wt = self.velocity
            losses.append(loss)
        return np.sum(losses) / len(losses)

    def softmax_equation(self, scores):
        """
        It calculates a softmax probability
        :param scores: A matrix(wt * input sample)
        :return: softmax probability
        """
        scores = np.max(scores)
        prob = (np.exp(scores).T / np.sum(np.exp(scores), axis=1)).T
        return prob

    def compute_loss(self, x, y_matrix):
        """
        It calculates a cross-entropy loss with regularization loss and gradient to update the weights.
        :param x: An input sample
        :param y_matrix: Label as one-hot encoding
        :return:
        """
        num_of_samples = x.shape[0]
        scores = np.dot(x, self.wt)
        prob = self.softmax_equation(scores)

        loss = -np.log(np.max(prob)) * y_matrix
        reg_loss = (1 / 2) * self.regularization_strength * np.sum(
            self.wt * self.wt)
        total_loss = (np.sum(loss) / num_of_samples) + reg_loss
        grad = ((-1 / num_of_samples) * np.dot(x.T, (y_matrix - prob))) + (
                self.regularization_strength * self.wt)
        return total_loss, grad

    def mean_accuracy(self, x, y):
        """
        It calculates mean-per class accuracy
        :param x: Input sample
        :param y: label sample
        :return: mean-per class accuracy
        """
        pred_y = self.predict(x)
        pred_y = pred_y.reshape((-1, 1))  # convert to column vector
        return np.mean(np.equal(y, pred_y))

    def predict(self, x):
        """
        It predict the label based on input sample and a model
        :param x: Input sample
        :return: predicted label
        """
        return np.argmax(x.dot(self.wt), 1)

    def one_hot_encoding(self, y, num_of_classes):
        """
        Convert a vector into one-hot encoding matrix where that particular column value is 1 and rest 0 for that row.
        :param y: Label vector
        :param num_of_classes: Number of unique labels
        :return: one-hot encoding matrix
        """
        y = np.asarray(y, dtype='int32')
        if len(y) > 1:
            y = y.reshape(-1)
        if not num_of_classes:
            num_of_classes = np.max(y) + 1
        y_matrix = np.zeros((len(y), num_of_classes))
        y_matrix[np.arange(len(y))] = 1
        return y_matrix

    def read_data(self, filename):
        """
        Read data from file and divide into input sample and a label.
        :param filename: name of a file
        :return: input sample and label
        """
        data = pd.read_csv(filename)
        X = data["X"].to_numpy()
        y = data["Y"].to_numpy()
        return X, y

    def train_model(self):
        """
        Main method
        """
        train_x, train_y = self.read_data(TRAIN_FILENAME)  # Training data
        test_x, test_y = self.read_data(TEST_FILENAME)  # Testing data

        parser = argparse.ArgumentParser()
        parser.add_argument("-e", "--epochs", dest="epochs", default=1000,
                            type=int, help="Number of epochs")
        parser.add_argument("-lr", "--learning_rate", dest="learning_rate", default=0.07,
                            type=float, help="Learning rate or step size")
        parser.add_argument("-bs", "--batch_size", dest="batch_size", default=10,
                            type=int, help="Number of sample in mini-batches")
        parser.add_argument("-r", "--regularization_strength", dest="regularization_strength", default=0.001,
                            type=float, help="L2 weight decay regularization lambda value")
        parser.add_argument("-m", "--momentum", dest="momentum", default=0.05,
                            type=float, help="A momentum value")

        args = parser.parse_args()

        print(
            "Epochs: {} | Learning Rate: {} | Batch Size: {} | Regularization Strength: {} | "
            "Momentum: {} |".format(
                args.epochs,
                args.learning_rate,
                args.batch_size,
                args.regularization_strength,
                args.momentum
            ))

        epochs = int(args.epochs)
        learning_rate = float(args.learning_rate)
        batch_size = int(args.batch_size)
        regularization_strength = int(args.regularization_strength)
        momentum = int(args.momentum)

        sm = HomepageRecommendation(epochs=epochs, learning_rate=learning_rate, batch_size=batch_size,
                                    regularization_strength=regularization_strength, momentum=momentum)
        train_losses, test_losses, train_acc, test_acc = sm.train(train_x, train_y, test_x, test_y)  # Train a network
        print("train_acc :", train_acc, "and test_acc:", test_acc)
        return train_losses, test_losses, train_acc, test_acc


if __name__ == '__main__':
    HomepageRecommendation(epochs=1000, learning_rate=0.07, batch_size=10,
                                    regularization_strength=0.001, momentum=0.05).train_model()