from typing import ClassVar

import ast
from operator import itemgetter
from datetime import datetime
import pandas as pd
from graphdb import GraphDb, GraphDbConnection
from ingestor.user_rating.constants import IMPLICIT_RATING, WEIGHTED_RATING
from pandas import DataFrame
from sklearn.preprocessing import MinMaxScaler

from offline_results.common.constants import USER_LABEL, USER_CONTENT_RELATIONSHIP_LABEL, CONTENT_ID, \
    CREATED_ON, VIEW_HISTORY, DURATION, CUSTOMER_ID, INNER, HAS_RATING, RATING_COUNT, RATING_SUM, \
    RATING_AVERAGE, SCALED_DURATION, SCALED_WEIGHTED_RATING, POPULAR_SCORE, RC_CREATED_ON


def get_recent_viewed_date_and_duration(history):
    history_data = history[1][VIEW_HISTORY]
    list_history_data = ast.literal_eval(history_data)
    list_history_data = sorted(list_history_data, key=itemgetter(CREATED_ON), reverse=True)
    recent_date = list_history_data[0][CREATED_ON]
    recent_date = datetime.strptime(recent_date, '%Y/%m/%d').date()

    total_duration = 0
    for history in list_history_data:
        duration_content = history[DURATION]
        total_duration += duration_content
    return recent_date, total_duration


class PopularModel:
    def __init__(
            self,
            connection_class: GraphDbConnection
    ):
        self.graph = GraphDb.from_connection(connection_class)

    @classmethod
    def from_connection_uri(
            cls,
            connection_uri: str
    ) -> ClassVar:
        """Create new object based on connection uri
        :param connection_uri: string connection uri
        :return: object class
        """
        return cls(GraphDbConnection.from_uri(connection_uri, connection_uri))

    @classmethod
    def from_connection_class(
            cls,
            connection_class: GraphDbConnection
    ) -> ClassVar:
        """Define new class based on object connection
        :param connection_class: object connection class
        :return: object class
        """
        return cls(connection_class)

    @staticmethod
    def fetch_required_data(graph):
        implicit_rating_queries = graph.custom_query(
            f''' g.V().hasLabel('{USER_LABEL}').outE().hasLabel('{HAS_RATING}').inV().path().by(elementMap())''',
            payload={USER_LABEL: USER_LABEL,
                     HAS_RATING: HAS_RATING}
        )
        user_rating_network_df = DataFrame()
        for query in implicit_rating_queries:
            for relation in query:
                customer_id = relation[0][CUSTOMER_ID]
                implicit_rating = relation[1][IMPLICIT_RATING]
                content_id = relation[2][CONTENT_ID]
                user_rating_data = pd.DataFrame([{
                    CUSTOMER_ID: customer_id,
                    CONTENT_ID: content_id,
                    IMPLICIT_RATING: implicit_rating
                }])
                user_rating_network_df = pd.concat([user_rating_network_df, user_rating_data], axis=0)

        viewed_duration_queries = graph.custom_query(
            f'''g.V().hasLabel('{USER_LABEL}').outE().hasLabel('{USER_CONTENT_RELATIONSHIP_LABEL}').inV().path().by(elementMap())''',
            payload={USER_LABEL: USER_LABEL,
                     USER_CONTENT_RELATIONSHIP_LABEL: USER_CONTENT_RELATIONSHIP_LABEL}
        )
        user_content_network_df = DataFrame()
        for query in viewed_duration_queries:
            for relation in query:
                customer_id = relation[0][CUSTOMER_ID]
                recent_viewed_date_data, total_duration = get_recent_viewed_date_and_duration(relation)
                content_id = relation[2][CONTENT_ID]
                user_content_data = pd.DataFrame([{
                    CUSTOMER_ID: customer_id,
                    CONTENT_ID: content_id,
                    DURATION: total_duration
                }])
                user_content_network_df = pd.concat([user_content_network_df, user_content_data], axis=0)
        final_df = user_content_network_df.merge(user_rating_network_df, how=INNER,
                                                 left_on=[CONTENT_ID, CUSTOMER_ID],
                                                 right_on=[CONTENT_ID, CUSTOMER_ID])
        tv_channel_list = [2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 17, 22, 38, 46, 48, 52, 55, 60, 62, 64, 2374, 2440,
                           2449, 2770, 2755, 2830, 2833, 2901, 2909, 2933, 6209, 6537, 6545, 6549, 6553, 7426, 7432,
                           7414, 19881]
        final_df = final_df[~final_df[CONTENT_ID].isin(tv_channel_list)].reset_index(drop=True)
        return final_df

    @staticmethod
    def preprocessed_required_data(data):
        rating_count_df = data.groupby([CONTENT_ID]).size().to_frame(RATING_COUNT)
        rating_sum_df = data.groupby(CONTENT_ID).sum().rename(columns={IMPLICIT_RATING: RATING_SUM})
        prepared_df = rating_sum_df.merge(rating_count_df, how=INNER, left_index=True, right_index=True)
        prepared_df[RATING_AVERAGE] = prepared_df[RATING_SUM] / prepared_df[RATING_COUNT]

        return prepared_df

    @staticmethod
    def get_popular_content(graph):
        required_data = PopularModel.fetch_required_data(graph)
        prepared_data = PopularModel.preprocessed_required_data(required_data)

        C = prepared_data[RATING_AVERAGE].mean()
        m = prepared_data[RATING_COUNT].quantile(0.3)
        listed_content_df = prepared_data.loc[prepared_data[RATING_COUNT] > m]

        def weighted_rating_formula(prepared_data, m=m, C=C):
            v = prepared_data[RATING_COUNT]
            R = prepared_data[RATING_AVERAGE]

            return (v / (v + m) * R) + (m / (m + v) * C)

        listed_content_df[WEIGHTED_RATING] = listed_content_df.apply(
            weighted_rating_formula, axis=1
        )
        listed_content_df = listed_content_df.sort_values(by=WEIGHTED_RATING, ascending=False)

        selected_df = listed_content_df[[DURATION, WEIGHTED_RATING]]
        scaled_features_df = MinMaxScaler().fit_transform(selected_df.values)
        scaled_features_df = pd.DataFrame(scaled_features_df, index=selected_df.index, columns=selected_df.columns)\
            .rename(columns={DURATION: SCALED_DURATION, WEIGHTED_RATING: SCALED_WEIGHTED_RATING})
        listed_content_df = listed_content_df.merge(scaled_features_df, how=INNER, left_index=True, right_index=True)
        listed_content_df[POPULAR_SCORE] = (listed_content_df[SCALED_DURATION]*0.5) + \
                                           (listed_content_df[SCALED_WEIGHTED_RATING]*0.5)
        listed_content_df = listed_content_df.sort_values(by=POPULAR_SCORE, ascending=False)
        listed_content_df[RC_CREATED_ON] = datetime.today().strftime("%Y-%m-%d %H:%M:%S")
        popular_content_df = listed_content_df[[RC_CREATED_ON]].rename_axis(CONTENT_ID).reset_index()
        return popular_content_df
