from typing import ClassVar

import pandas as pd
from graphdb import GraphDb, GraphDbConnection
from sklearn.preprocessing import MinMaxScaler
from datetime import datetime

from offline_results.common.constants import CONTENT_ID, VIEW_COUNT, DURATION, UBD_DAYOFWEEK, SCALED_DURATION, \
    SCALED_VIEW_COUNT, INNER, TRENDING_SCORE, UBD_DAYOFYEAR, RC_CREATED_ON, UBD_CREATED_ON, SEASON, WEEKEND, TGIF, \
    NEW_YEAR, CHRISTMAS, VALENTINE

from offline_results.recommendation.homepage_recommendation.trending.trending_model import TrendingModel


class SeasonalModel:
    def __init__(
            self,
            connection_class: GraphDbConnection
    ):
        self.graph = GraphDb.from_connection(connection_class)

    @classmethod
    def from_connection_uri(
            cls,
            connection_uri: str
    ) -> ClassVar:
        """Create new object based on connection uri
        :param connection_uri: string connection uri
        :return: object class
        """
        return cls(GraphDbConnection.from_uri(connection_uri, connection_uri))

    @classmethod
    def from_connection_class(
            cls,
            connection_class: GraphDbConnection
    ) -> ClassVar:
        """Define new class based on object connection
        :param connection_class: object connection class
        :return: object class
        """
        return cls(connection_class)

    @staticmethod
    def get_specific_seasons(history_content_df_vod):
        history_content_df_vod[UBD_DAYOFYEAR] = history_content_df_vod[UBD_CREATED_ON].dt.dayofyear

        weekend_season = history_content_df_vod[history_content_df_vod[UBD_DAYOFWEEK].isin([5, 6])]\
            .reset_index(drop=True)
        tgif_season = history_content_df_vod[history_content_df_vod[UBD_DAYOFWEEK].isin([4])]\
            .reset_index(drop=True)
        new_year_season = history_content_df_vod[history_content_df_vod[UBD_DAYOFYEAR].isin([1, 365])]\
            .reset_index(drop=True)
        christmas_season = history_content_df_vod[history_content_df_vod[UBD_DAYOFYEAR].isin([358, 359, 360])]\
            .reset_index(drop=True)
        valentine_season = history_content_df_vod[history_content_df_vod[UBD_DAYOFYEAR].isin([45])]\
            .reset_index(drop=True)

        df_season_list = [weekend_season, tgif_season, new_year_season, christmas_season, valentine_season]
        return df_season_list

    @staticmethod
    def get_seasonal_content(graph):
        history_content_df, history_content_df_vod = TrendingModel.fetch_historical_data(graph)
        df_season_list = SeasonalModel.get_specific_seasons(history_content_df_vod)

        seasonal_content_df = pd.DataFrame()
        for idx, df in enumerate(df_season_list):
            df_view_count = df.groupby([CONTENT_ID]).size().to_frame(VIEW_COUNT)
            df_total_duration_content = df.groupby([CONTENT_ID]).sum()
            df_total_duration_content = df_total_duration_content[[DURATION]]
            final_recommendation_df = pd.merge(df_total_duration_content, df_view_count,
                                               left_index=True, right_index=True).sort_values(by=[DURATION, VIEW_COUNT],
                                                                                              ascending=False)
            scaled_features_df = MinMaxScaler().fit_transform(final_recommendation_df.values)
            scaled_features_df = pd.DataFrame(scaled_features_df, index=final_recommendation_df.index,
                                              columns=final_recommendation_df.columns).rename(
                columns={DURATION: SCALED_DURATION, VIEW_COUNT: SCALED_VIEW_COUNT}
            )
            final_recommendation_df = final_recommendation_df.merge(scaled_features_df, how=INNER,
                                                                    left_index=True, right_index=True)
            final_recommendation_df[TRENDING_SCORE] = (final_recommendation_df[SCALED_DURATION] * 0.7) + \
                                                      (final_recommendation_df[SCALED_VIEW_COUNT] * 0.3)
            final_recommendation_df = final_recommendation_df.sort_values(by=TRENDING_SCORE, ascending=False)
            final_recommendation_df[RC_CREATED_ON] = datetime.today().strftime("%Y-%m-%d %H:%M:%S")
            final_recommendation_df = final_recommendation_df[[RC_CREATED_ON]].rename_axis(CONTENT_ID).reset_index()
            season_map = {0: WEEKEND, 1: TGIF, 2: NEW_YEAR, 3: CHRISTMAS, 4: VALENTINE}
            final_recommendation_df[SEASON] = season_map[idx]
            seasonal_content_df = pd.concat([seasonal_content_df, final_recommendation_df], axis=0)\
                .reset_index(drop=True)

        return seasonal_content_df
